/*
   bitset.c -- PIKT piktc bitset routines

   $Header: /var/cvs/PIKT/pikt/src/piktc/bitset.c,v 1.27 2005/01/04 13:48:31 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998,1999,2000,2001,2002,2003,2004,2005 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org
*/

#include "../lib/common.h"
#include "piktc.h"

unsigned long *ghbs;
unsigned long *gdbs, *gubs;
unsigned long *hbsfull, *hbsempty;
unsigned long *dbsfull, *dbsempty;

/* empty a bitset (turn off all bits)
 */
void
empty (unsigned long *bs, size_t bslongs)
{
	register int i;

	for (i=0; i<bslongs; i++)
		*(bs+i) = 0UL;
	return;
}

/* fill a bitset (turn on all bits)
 */
void
fill (unsigned long *bs, size_t bslongs)
{
	register int i;

	for (i=0; i<bslongs; i++)
		*(bs+i) = ~0UL;
	return;
}

/* union of two bitsets;
   returns a calloc()'ed bs that should eventually be free()'ed
 */
unsigned long *
bsunion (unsigned long *bs1, unsigned long *bs2, size_t bslongs)
{
	static char fnnam[] = "bsunion";
	register int i;
	unsigned long *bs;

	bs = (unsigned long *) mycalloc(bslongs, sizeof(unsigned long), fnnam);
	for (i=0; i<bslongs; i++)
		*(bs+i) = *(bs1+i) | *(bs2+i);
	return bs;
}

/* intersection of two bitsets;
   returns a calloc()'ed bs that should eventually be free()'ed
 */
unsigned long *
bsintersection (unsigned long *bs1, unsigned long *bs2, size_t bslongs)
{
	static char fnnam[] = "bsintersection";
	register int i;
	unsigned long *bs;

	bs = (unsigned long *) mycalloc(bslongs, sizeof(unsigned long), fnnam);
	for (i=0; i<bslongs; i++)
		*(bs+i) = *(bs1+i) & *(bs2+i);
	return bs;
}

/* complement of a bitset;
   returns a calloc()'ed bs that should eventually be free()'ed
 */
unsigned long *
bscomplement (unsigned long *bs1, size_t bslongs)
{
	static char fnnam[] = "bscomplement";
	register int i;
	unsigned long *bs;

	bs = (unsigned long *) mycalloc(bslongs, sizeof(unsigned long), fnnam);
	for (i=0; i<bslongs; i++) {
		*(bs+i) = ~*(bs1+i);
	}
	return bs;
}

/* is item (the second bitset--a sequence of 0's and just a single 1)
   a member of a set (the first bitset--a sequence of 0's and 1's)
 */
int
bsmember (unsigned long *bs1, unsigned long *bs2, size_t bslongs)
{
	register int i;
	static unsigned long *bs;

	bs = bsintersection(bs1, bs2, bslongs);
	for (i=0; i<bslongs; i++)
		if (*(bs+i)) {
			free(bs);
			return TRUE;
		}
	free(bs);
	return FALSE;
}

/* is bs2 a subset of bs1?  if the intersection of bs2 with bs1's complement
   is not the empty set, return false, else return true
 */
int
bssubset (unsigned long *bs1, unsigned long *bs2, size_t bslongs)
{
	register int i;
	static unsigned long *bs;

	bs = bscomplement(bs1, bslongs);
	for (i=0; i<bslongs; i++) {
		if (*(bs+i) & *(bs2+i)) {
			free(bs);
			return FALSE;
		}
	}
	free(bs);
	return TRUE;
}

/* is even just a single bit turned on in the given bit set?
 * /
unsigned long
defined (unsigned long *bs, size_t bslongs)
{
	register int i;
	unsigned long d;

	d = 0UL;
	for (i=0; i<bslongs; i++)
		d |= *(bs+i);
	return d;
}
*/

/* copy one hostset to another
 */
void
bscopy (unsigned long *bs1, unsigned long *bs2, size_t bslongs)
{
	register int i;

	for (i=0; i<bslongs; i++)
		*(bs1+i) = *(bs2+i);
	return;
}

/* duplicate a bitset, returning a pointer to the duplicate;
   returns a calloc()'ed bs that should eventually be free()'ed
 */
unsigned long *
bsdup (unsigned long *bs2, size_t bslongs)
{
	static char fnnam[] = "bsdup";
	register int i;
	unsigned long *bs;

	bs = (unsigned long *) mycalloc(bslongs, sizeof(unsigned long), fnnam);
	for (i=0; i<bslongs; i++)
		*(bs+i) = *(bs2+i);
	return bs;
}

/* get the bs (bitset) for the specified ident
 */
unsigned long *
getbs (char *identname, int type, int permit_unknown_ident)
{
	static char fnnam[] = "getbs";
	ident *idp;

	switch (type) {
		case OPS:
		case HST:
		case GRP:
		case ALS:
			idp = lookupident(identname, SYSCFG);
			if (idp)
				return idp->bs;
			else {
				if (permit_unknown_ident) {
					if (isverboselog)
						logreturn(PIKTLOG, LOG_WARNING, ERRID_HOST_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "%s, at or near line %d: os, host, group, or alias %s unknown", cfgfilecurr, lineno, identname);
					return hbsempty;
				} else {
					logexit(OUTPUT, LOG_ERR, ERRID_HOST_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "%s, at or near line %d: os, host, group, or alias %s unknown", cfgfilecurr, lineno, identname);
				}
			}
			break;
		case DEF:
			idp = lookupident(identname, DEFCFG);
			if (idp) {
				return idp->bs;
			} else {
				logexit(OUTPUT, LOG_ERR, ERRID_DEFINE_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "%s, at or near line %d: define %s unknown", cfgfilecurr, lineno, identname);
			}
			break;
		default:
			return (unsigned long *) NULL;
			break;
	}
	return (unsigned long *) NULL;	/* should never be reached */
}

/* apply the bitset tag to the current config line, also its line number,
   cfg file, and length, and whether it should be macro expanded; mark it
   for processing by default
 */
void
tag (int lno, char *file, unsigned long *bs, size_t bslongs, int domcr, int doproc, char *line)
{
	static char fnnam[] = "tag";
	cfgln *nclp;

	nclp = (cfgln *) mymalloc(sizeof(cfgln), fnnam);
	nclp->next = (cfgln *) NULL;
	nclp->lineno = lno;
	if (clplast) {
		if (strcmp(clplast->file, file))
			nclp->file = mystrdup(file, fnnam);
		else
			nclp->file = clplast->file;
	} else
		nclp->file = mystrdup(file, fnnam);
	nclp->bs = bsdup(bs, bslongs);
	nclp->line = line;
	nclp->len = strlen(line);
	nclp->domcr = domcr;
	nclp->doproc = doproc;
	if (! configs[cfgcurr]) {
		configs[cfgcurr] = nclp;
	} else {
		clplast->next = nclp;
	}
	clplast = nclp;
}

/* insert the current ident into the global bitset
 */
ident *
insert (char *identname, int type)
{
	static char fnnam[] = "insert";
	ident *idp;
	register int i;
	
	switch (type) {
		case HST:	/* or OPS, GRP, or ALS */
			idp = lookupident(identname, SYSCFG);
			if (idp) {
				for (i=0; i<hbslongs; i++)
					*(ghbs+i) |= *((idp->bs)+i);
			} else {
				logreturn(PIKTLOG, LOG_ERR, ERRID_IDENT_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "ident %s not registered", identname);
				bailout = TRUE;
			}
			return idp;
			break;
		case DEF:

			idp = lookupident(identname, DEFCFG);
			if (idp) {
				for (i=0; i<dbslongs; i++)
					*(gdbs+i) |= *((idp->bs)+i);
			} else {
				logreturn(PIKTLOG, LOG_ERR, ERRID_DEFINE_UNKNOWN, __FILE__, __LINE__, fnnam, FALSE, "define %s not registered", identname);
				bailout = TRUE;
			}
			return idp;
			break;
		default:
			return (ident *) NULL;
			break;
	}
}
