/*
   expr.c -- PIKT script expression routines

   $Header: /var/cvs/PIKT/pikt/src/piktd/expr.c,v 1.54 2005/01/04 13:49:04 berto Exp $

   This file is part of PIKT ("Problem Informant/Killer Tool").
   Copyright (C) 1998,1999,2000,2001,2002,2003,2004,2005 by Robert Osterlund.

   PIKT is a trademark of the University of Chicago.  All Rights Reserved.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.

   PIKT is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
   USA.

   Send problem reports, bouquets and brickbats to:  pikt-bugs@pikt.org
 */

#include "../lib/common.h"

#include <math.h>
#include <pwd.h>
#include <grp.h>

#include "piktd.h"

#define MAXTBUF		64

#define SECS_IN_HOUR	(60 * 60)
#define SECS_IN_DAY	(24 * 60 * 60)
#define SECS_IN_YEAR	(365 * 24 * 60 * 60)

int numcnt;
double numsum;
double nummax;
double nummin;
st *stats = (st *) NULL;

const char *monthnames[] = {
	"January", "February", "March", "April", "May", "June",
	"July", "August", "September", "October", "November", "December"
};

const char *daynames[] = {
	"Sunday", "Monday", "Tuesday", "Wednesday",
	"Thursday", "Friday", "Saturday"
};

/* return the current time value
 */
time_t
now (void)
{

	return time((time_t*)0);
}

/* return the time value as of last midnight--the number of seconds
   since the Unix epoch
 */
time_t
today (void)
{

	return time((time_t*)0);
}

/* return the time value for the given year, month & day--the number of
   seconds since the Unix epoch
 */
time_t
datevalue (int yyyy, int mm, int dd)
{
	static struct tm ts;
	time_t t;

	ts.tm_sec = 0;
	ts.tm_min = 0;
	ts.tm_hour = 0;
	ts.tm_mday = dd;
	ts.tm_mon = mm - 1;
	ts.tm_year = yyyy - 1900;
	ts.tm_isdst = 0;	/* initialization */
	t = mktime(&ts);
	return t;
}

/* return the time value for the given hour, minute & second--the number
   of seconds since midnight
 */
time_t
timevalue (int hr, int min, int sec)
{
	time_t t;

	t = 60*60*hr + 60*min + sec;
	return t;
}

/* return the second (0..59) of the given time value
 */
int
second (time_t t)
{
	char tbuf[MAXTBUF];
	struct tm *tp;

	tp = localtime(&t);
	strftime(tbuf, MAXTBUF, "%S", tp);
	return atoi(tbuf);
}

/* return the minute (0..59) of the given time value
 */
int
minute (time_t t)
{
	char tbuf[MAXTBUF];
	struct tm *tp;

	tp = localtime(&t);
	strftime(tbuf, MAXTBUF, "%M", tp);
	return atoi(tbuf);
}

/* return the hour (0..23) of the given time value
 */
int
hour (time_t t)
{
	char tbuf[MAXTBUF];
	struct tm *tp;

	tp = localtime(&t);
	strftime(tbuf, MAXTBUF, "%H", tp);
	return atoi(tbuf);
}

/* return the day (1..31) of the given time value
 */
int
day (time_t t)
{
	char tbuf[MAXTBUF];
	struct tm *tp;

	tp = localtime(&t);
	strftime(tbuf, MAXTBUF, "%d", tp);
	return atoi(tbuf);
}

/* return the month (1..12) of the given time value
 */
int
month (time_t t)
{
	char tbuf[MAXTBUF];
	struct tm *tp;

	tp = localtime(&t);
	strftime(tbuf, MAXTBUF, "%m", tp);
	return atoi(tbuf);
}

/* return the year (in YYYY form) of the given time value
 */
int
year (time_t t)
{
	char tbuf[MAXTBUF];
	struct tm *tp;

	tp = localtime(&t);
	strftime(tbuf, MAXTBUF, "%Y", tp);
	return atoi(tbuf);
}

/* return the weekday (0..6) of the given time value
 */
int
weekday (time_t t)
{
	char tbuf[MAXTBUF];
	struct tm *tp;

	tp = localtime(&t);
	strftime(tbuf, MAXTBUF, "%w", tp);
	return atoi(tbuf)+1;
}

/* return the yearday (1..366) of the given time value
 */
int
yearday (time_t t)
{
	char tbuf[MAXTBUF];
	struct tm *tp;

	tp = localtime(&t);
	strftime(tbuf, MAXTBUF, "%j", tp);
	return atoi(tbuf);
}

/* return the yearweek (1..53) of the given time value
 */
int
yearweek (time_t t)
{
	char tbuf[MAXTBUF];
	struct tm *tp;

	tp = localtime(&t);
	strftime(tbuf, MAXTBUF, "%U", tp);
	return atoi(tbuf);
}

/* return "AM" or "PM" for the given time value
 */
char *
ampm (time_t t)
{
	static char tbuf[MAXTBUF];
	struct tm *tp;

	tp = localtime(&t);
	strftime(tbuf, MAXTBUF, "%p", tp);
	return tbuf;
}

/* given a month string like "January", "February", ... (or their three-char
   equivalents "Jan", "Feb", ...), return the month number (1..12); this
   function must be insensitive to case!
 */
int
monthnumber (char *monnam)
{
	register int i;

	for (i=0; i<12; i++) {
		if (! strcasecmp(monthnames[i],monnam) || ((strlen(monnam) == 3) && ! strncasecmp(monthnames[i],monnam,3)))
			return i+1;
	}
	return -1;
}

/* given a number from 1 to 12, return the corresponding month name
   "January", "February", ...; the input number is not error-checked
   here, but rather at the point of monthname()'s invocation in exprparse()
 */
const char *
monthname (int monnum)
{

	return monthnames[monnum-1];
}

/* given a day string like "Monday", "Tuesday", ... (or their three-char
   equivalents "Mon", "Tue", ...), return the day number (1..7)
 */
int
daynumber (char *daynam)
{
	register int i;

	for (i=0; i<7; i++) {
		if (! strcmp(daynames[i],daynam) || ((strlen(daynam) == 3) && ! strncmp(daynames[i],daynam,3)))
			return i+1;
	}
	return ERROR;
}

/* given a number from 1 to 7, return the corresponding day name
   "Monday", "Tuesday", ...; the input number is not error-checked
   here, but rather at the point of dayname()'s invocation in exprparse()
 */
const char *
dayname (int daynum)
{

	return daynames[daynum-1];
}

/* return the number of days between two two dates (time values)
 */
int
daysbetween (time_t begin, time_t end)
{

	return (int) (end - begin)/SECS_IN_DAY;
}

/* given a month, day & time, or possibly year instead of time,
   return the number of days since then
 */
int
age (char *mon, char *dy, char *tim)
{
	int m, d, y;
	int mnthen, mnnow;
	time_t tnow;

	tnow = now();
	m = monthnumber(mon);
	d = atoi(dy);
	if (tim[2] == ':') {
		y = year(tnow);
		mnthen = monthnumber(mon);
		mnnow = month(tnow);
		if (mnnow - mnthen < 0)
			y--;
	} else	/* "time" is a four-digit year */
		y = atoi(tim);
	return daysbetween(datevalue(y,m,d), tnow);
}

/* given a string, return the specified number of chars beginning at the
   specified position; the invoking exprparse() ensures that the given
   number and position are non-negative
   e.g., substring("this is a test", 4, 4) --> "s is"
         substring("this is a test", 1, 20) --> "this is a test"
         substring("this is a test", 20, 1) --> ""
 */
char *
substring (char *s, int b, size_t e)
{
	static char fnnam[] = "substring";
	int slen;
	char *r;

	slen = strlen(s);
	if ((b > slen) || (b == 0)) {
		r = mystrdup(nullstring, fnnam);
	} else {
		r = mystrdup(s+b-1, fnnam);
		if ((b-1+e) < slen)
			*(r+e) = '\0';
	}
	return r;
}

/* chop the specified number of chars off the right end of the given string;
   e.g., chop("this is a test", 7) --> "this is"
         chop("this is a test", 0) --> "this is a test"
         chop("this is a test", 20) --> ""
 */
char *
chop (char *s, int n)
{
	static char fnnam[] = "chop";
	char *r;
	register int i, j, k;
	register char c;

	r = mystrdup(s, fnnam);
	i = j = 0;
	while ((c = *(s+i))) {
		*(r+j) = c;
		i++;
		j++;
	}
	if ((k = (i-n)) < 0)
		*r = '\0';
	else
		*(r+k) = '\0';
	return r;
}

/* trim white space off the beginning and/or end of the given string
   e.g., trim("  test  ", BOTH)  --> "test"
         trim("  test  ", LEFT)  --> "test  "
         trim("  test  ", RIGHT) --> "  test"
 */
signed char *
trim (char *s, int type)
{
	static char fnnam[] = "trim";
	signed char *r;
	register int i, j;
	register char c;

	r = mystrdup(s, fnnam);
	i = 0;
	j = 0;
	while ((c = *(s+i))) {
		switch (type) {
			case LEFT:
			case BOTH:
				if (((c != ' ') && (c != '\t')) || j) {
					*(r+j) = c;
					j++;
				}
				break;
			case RIGHT:
				/* no need, mystrdup() above did this: */
				/* *(r+j) = c; */
				j++;
				break;
		}
		i++;
	}
	*(r+j) = '\0';
	if (j)
		j--;
	if ((type == RIGHT) || (type == BOTH))
		while ((c = *(r+j)) && ((c == ' ') || (c == '\t'))) {
			*(r+j) = '\0';
			if (--j < 0)
				break;
		}
	return r;
}

/* find a needle string in a haystack string, returning its position;
   e.g., &search("this is a test", "is") --> 3
   with an optional third argument, begin the search at that position;
   e.g., &search("this is a test", "is", 4) --> 6
   if the needle is not found, or the needle and/or haystack are the null
   string, return 0
   if the position is zero or less, return ERROR
   if the position is extends beyond the end of the haystack, return 0
 */
int
search (char *h, char *n, int p)
{
	char *cp;

	if (p <= 0)
		return ERROR;
	if (p > strlen(h))
		return 0;
	if (! *h || ! *n)
		return 0;
	cp = strstr(h+p-1,n);
	if (cp)
		return (int) (cp - h) + 1;
	else
		return 0;
}

/* do a reverse search for a needle in a haystack, returning the position of
   the found needle;
   e.g., &rsearch("this is a test", "is") --> 6
         &rsearch("this is a test", "t") --> 14
         &rsearch("this is a test", "foo") --> 0
 */
int
rsearch (char *h, char *n)
{
	static char *hr, *nr;
	int pos, r;

	hr = reverse(h);
	nr = reverse(n);
	if ((pos = search(hr, nr, 1)) != 0)
		r = strlen(hr) - (pos-1) - strlen(nr) + 1;
	else
		r = 0;
	free(hr);
	free(nr);
	return r;
}

/* reverse the given string
   e.g., reverse("this is a test") --> "tset a si siht"
 */
char *
reverse (char *s)
{
	static char fnnam[] = "reverse";
	char *r;
	register int i, j;
	register int sl;

	r = mystrdup(s, fnnam);
	i = 0;
	j = sl = strlen(s);
	*(r+j--) = '\0';
	for (i=0; i<sl; i++, j--) {
		*(r+j) = *(s+i);
	}
	return r;
}

/* return TRUE (non zero) or FALSE (zero) depending on whether the given
   file path meets certain conditions
 */
int
filestat (char *path, int type)
{
	static char fnnam[] = "filestat";
	struct stat fbuf;

	if (lstat(path, &fbuf) != 0) {
		if (isverboselog)
			logreturn(PIKTLOG, LOG_WARNING, ERRID_FILE_STAT_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, lstat() error for %s", ALMNAM, path);
		return 0;
	} else {
		switch (type) {
			case EXIST:
				return 1;
				break;
			case ZEROLEN:
				return fbuf.st_size ? FALSE : TRUE;
				break;
			case REGFILE:
				return S_ISREG(fbuf.st_mode);
				break;
			case DIRECTORY:
				return S_ISDIR(fbuf.st_mode);
				break;
			case CHARSPEC:
				return S_ISCHR(fbuf.st_mode);
				break;
			case BLKSPEC:
				return S_ISBLK(fbuf.st_mode);
				break;
			case PIPE:
				return S_ISFIFO(fbuf.st_mode);
				break;
			case SYMLINK:
				return S_ISLNK(fbuf.st_mode);
				break;
			case SOCKET:
				return S_ISSOCK(fbuf.st_mode);
				break;
			default:
				logreturn(PIKTLOG, LOG_ERR, ERRID_FILE_STAT_TYPE_INVALID, __FILE__, __LINE__, fnnam, FALSE, "in %s, unknown filestat type %d", ALMNAM, type);
				return ERROR;
				break;
		}
	}
}

/* given a string, return the ASCII value of the first char; e.g.,
   "ASCII" --> 65; the empty string returns 0
 */
int
code (char *string)
{

	return (int) *string;
}

/* given an int between 1 & 127, return its char equivalent; e.g., 65 --> A
 */
char *
my_char (int cd)
{
	static char fnnam[] = "my_char";
	static char s[2];
	char *r;

	if ((cd >= 0) && (cd <= 127))
		sprintf(s,"%c",cd);
	else
		s[0] = '\0';
	r = mystrdup(s, fnnam);
	return r;
}

/* round up to nearest integer;
   e.g., &ceiling(2.1) --> 3
   ceiling(-2.1) --> -2
 */
int
ceiling (double number)
{

	return -integer(-number);
}

/* round down to nearest integer;
   e.g., &integer(2.5) --> 2
   &integer(-2.5) --> -3
 */
int
integer (double number)
{

	return (isint(number) || (number >= 0)) ? (int) number : (int) (number - 1) ;
}

/* round up or down to nearest integer; 
   my_round(2.5) --> 3
   my_round(-2.5) --> -3
*/
int
my_round (double number)
{

	if (number >= 0)
		return (int) (number+0.5);
	else
		return (int) (number-0.5);
}

/* truncate number to integer by dropping fractional part;
   e.g., &trunc(2.5) --> 2
   &trunc(-2.5) --> -2
 */
int
my_trunc (double number)
{

	return (int) number;
}

/* determine whether a double is an integer or not
 */
int
isint (double n)
{

	if (n >= 0) {
		if ((ceil(n) - n) < 0.001)	/* any number < 1 will do here */
			return TRUE;
		else
			return FALSE;
	} else {
		if ((n - floor(n)) < 0.001)	/* any number < 1 will do here */
			return TRUE;
		else
			return FALSE;
	}
}

/* given a floating point number, returns a string formatted to the given
   decimal places; e.g., &fixed(3.1416, 2) --> "3.14"
 */
char *
fixed (double number, int decimals)
{
	static char fnnam[] = "fixed";
	char s[MAXLINE+1];
	char fmt [16+1];
	char *r;

	if (decimals < 0) {
		r = mystrdup(err, fnnam);
		return r;
	}
	sprintf(fmt, "%%.%df", decimals);
	sprintf(s, fmt, number);
	r = mystrdup(s, fnnam);
	return r;
}

/* return the given string all in lower case
   e.g., lower("TEST") --> "test"
 */
char *
lower (char *s)
{
	static char fnnam[] = "lower";
	char *r;
	register int i;
	register char c;

	r = mystrdup(s, fnnam);
	i = 0;
	while ((c = *(s+i))) {
		*(r+i) = tolower(c);
		i++;
	}
	*(r+i) = '\0';
	return r;
}

/* return the given string all in upper case
   e.g., lower("test") --> "TEST"
 */
char *
upper (char *s)
{
	static char fnnam[] = "upper";
	char *r;
	register int i;
	register char c;

	r = mystrdup(s, fnnam);
	i = 0;
	while ((c = *(s+i))) {
		*(r+i) = toupper(c);
		i++;
	}
	*(r+i) = '\0';
	return r;
}

/* return the given string with the first letter capitalized
   e.g., proper("test") --> "Test"
 */
char *
proper (char *s)
{
	static char fnnam[] = "proper";
	char *r;
	register int i;
	register char c;
	int spaceprev, spacecurr;

	r = mystrdup(s, fnnam);
	spaceprev = 1;
	i = 0;
	while ((c = *(s+i))) {
		spacecurr = isspace((int)c);
		if (! spacecurr && spaceprev)
			*(r+i) = toupper(c);
		else
			*(r+i) = c;
		spaceprev = spacecurr;
		i++;
	}
	*(r+i) = '\0';
	return r;
}

/* repeat a string n number of times;
   e.g., &repeat("t",4) --> "tttt"
         &repeat("this",4) --> "thisthisthisthis"
         &repeat("A",0) --> ""
 */
char *
repeat (char *s, int n)
{
	static char fnnam[] = "repeat";
	char *r;
	register int i;

	r = (char *) mymalloc(n*strlen(s)+1, fnnam);
	*r = '\0';
	i = 0;
	while (i < n) {
		strcat(r, s);
		i++;
	}
	return r;
}

/* in a target string, replace chars, beginning at a given position and
   ranging through a given number of chars, with a substitution string;
   e.g., &replace("this is a test", "at", 3, 2) --> "that is a test"
         &replace("this is a test", "at", 3, 20) --> "that"
         &replace("this is a test", "at", 3, 0) --> "this is a test"
         &replace("this is a test", "", 3, 0) --> "this is a test"
         &replace("this is a test", "", 3, 20) --> "th"
         &replace("", "at", 1, 2) --> ""
 */
char *
replace (char *old, char *new, int pos, int chars)
{
	static char fnnam[] = "replace";
	char *r;
	int oldlen, newlen, len;

	if (chars == 0) {
		r = mystrdup(old, fnnam);
		return r;
	}
	oldlen = strlen(old);
	newlen = strlen(new);
	if (pos > oldlen) {
		r = mystrdup(old, fnnam);
		return r;
	}
	r = (char *) mymalloc(strlen(old)+strlen(new)+1, fnnam);
	strncpy(r, old, (size_t)(pos-1));
	len = pos-1;
	*(r+pos-1) = '\0';
	if (*new) {	/* i.e., not null string */
		strcat(r, new);
		len += newlen;
	}
	if (pos-1+chars < oldlen)
		strcat(r, old+pos-1+chars);
	return r;
}

/* in a target string, substitute some old chars with some new chars;
   if specified, substitute only the old chars in the specified instance,
   else substitute all old chars
   e.g., &substitute("this is a test", "t", "X") --> "Xhis is a XesX"
         &substitute("this is a test", "t", "X", 2) --> "this is a Xest"
         &substitute("this is a test", "t", "") --> "his is a es"
         &substitute("this is a test", "is", "at") --> "that at a test"
   we allow for a target string of up to MAXBUF size; we return the err
   string if it threatens to go over this; although we could call realloc()
   to allow for really huge strings, MAXBUF is already large enough for
   most reasonable situations that implementing realloc() is really overkill;
   we're not trying to achieve perl-like universality here
 */
char *
substitute (char *s, char *old, char *new, int instance)
{
	static char fnnam[] = "substitute";
	char *r;
	char *needle, *haystack;
	register int oldlen, newlen, len;
	register int count;

	r = (char *) mymalloc(MAXBUF+1, fnnam);		/* arbitrarily large */
	*r = '\0';
	if (*old)
		oldlen = strlen(old);
	else
		return r;
	newlen = strlen(new);
	len = 0;
	count = 0;
	haystack = s;
	while ((needle = strstr(haystack, old))) {
		count++;
		if (len + (needle-haystack) > MAXBUF) {
			strcpy(r, err);
			return r;
		} else {
			strncat(r, haystack, (size_t)(needle-haystack));
			len += needle-haystack;
		}
		if ((count == instance) || (instance == 0)) {
			if (len + newlen > MAXBUF) {
				strcpy(r, err);
				return r;
			} else {
				strcat(r, new);
				len += newlen;
			}
		} else {
			if (len + oldlen > MAXBUF) {
				strcpy(r, err);
				return r;
			} else {
				strcat(r, old);
				len += oldlen;
			}
		}
		haystack = needle + oldlen;
	}
	if (len + strlen(haystack) > MAXBUF) {
		strcpy(r, err);
		return r;
	} else
		strcat(r, haystack);
	return r;
}

/* register the current statistic in the stats list; used by the statistic
   functions
 */
void
regstat (double num)
{
	static char fnnam[] = "regstat";
	static st *nsp;
	st *stcurr;

	numcnt++;
	numsum += num;
	if (numcnt == 1) {
		nummax = num;
		nummin = num;
	} else {
		if (num > nummax)
			nummax = num;
		if (num < nummin)
			nummin = num;
	}
	nsp = (st *) mymalloc(sizeof(st), fnnam);
	nsp->num = num;
	nsp->freq = 1;
	nsp->next = (st *) NULL;
	if (! stats) {
		stats = nsp;
		return;
	} else {
		if (num < stats->num) {
			nsp->next = stats;
			stats = nsp;
			return;
		} else if (num == stats->num) {
			stats->freq++;
			free(nsp);
			return;
		} else {
			stcurr = stats;
			while (stcurr) {
				if (! stcurr->next) {
					stcurr->next = nsp;
					return;
				} else if (num < stcurr->next->num) {
					nsp->next = stcurr->next;
					stcurr->next = nsp;
					return;
				} else if (num == stcurr->next->num) {
					stcurr->next->freq++;
					free(nsp);
					return;
				}
				stcurr = stcurr->next;
			}
		}
	}
	return;
}

/* garbage collect the no-longer-needed registered statistics
 */
void
freestats (void)
{
	static st *sp, *tsp;

	sp = stats;
	while (sp) {
		tsp = sp;
		sp = sp->next;
		free(tsp);
	}
	stats = (st *) NULL;
	return;
}

/* for a list of stat vals, return their median val
   e.g., with the stat list 1,5,1,5,2,4,3,3,3, median() --> 3
 */
double
median (void)
{
	st *sp;
	int mcnt;
	int cnt;

	mcnt = (int) ceiling((double)(numcnt/2.0));
	cnt = 0;
	sp = stats;
	while (sp) {
		cnt += sp->freq;
		if (cnt >= mcnt)
			break;
		sp = sp->next;
	}
	return sp->num;
}

/* for a list of stat vals, return their modal val
   e.g., with the stat list 1,5,1,5,2,4,3,3,3, mode() --> 3
 */
double
mode (void)
{
	st *sp;
	int maxfreq;
	double mnum;

	mnum = 0.0;	/* initialize; what else but zero? */
	maxfreq = 0;
	sp = stats;
	while (sp) {
		if (sp->freq > maxfreq) {
			maxfreq = sp->freq;
			mnum = sp->num;
		}
		sp = sp->next;
	}
	return mnum;
}

/* when invoked as $command(), return the first output line for the given
   command string; when invoked as #system(), return the exit value for
   the given command string
 */
char *
sys (signed char *cmd, int type)
{
	static char fnnam[] = "sys";
	FILE *infile;
	pid_t pid;
	char buf[MAXBUF+1];
	char *cmd2;
	unsigned int cmdlen;
	int i;
	char *r;

	if ((cmdlen = strlen(cmd)) > MAXCMD) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_COMMAND_STRING_LEN_MAX_EXCEEDED, __FILE__, __LINE__, fnnam, FALSE, "in %s, cmd exceeds %d char limit", ALMNAM, MAXCMD);
		r = mystrdup(err, fnnam);
		return r;
	} else if (cmdlen == 0) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_COMMAND_STRING_MISSING, __FILE__, __LINE__, fnnam, FALSE, "in %s, null cmd string", ALMNAM);
		r = mystrdup(err, fnnam);
		return r;
	}
	/* 26 for length of " >/dev/null 2>&1; echo $?" plus terminating
	   nullchar
	 */
	cmd2 = mycalloc(cmdlen+26, sizeof(char), fnnam);
	strncpy(cmd2, cmd, cmdlen);
	if (type == SYSTEM)
		strcat(cmd2, " >/dev/null 2>&1; echo $?");
	requote(cmd2);
	logreturn(PIKTLOG, LOG_INFO, ERRID_INFO, __FILE__, __LINE__, fnnam, FALSE, "in %s, command \"%s\"", ALMNAM, cmd2);
	/* for some strange reason, in later versions of AIX (confirmed in
	   4.3.3 running on a PowerPC) but not earlier versions, 'echo $?'
	   always gives 0; changing this to

		strcat(cmd2, " >/dev/null 2>&1 && echo 0 || echo 1");

	   yields the same results--invariably 0.  crazy!
	 */
	if ((infile = inputopen(cmd2, &pid)) == NULL) {
		logreturn(PIKTLOG, LOG_ERR, ERRID_INPUT_OPEN_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, inputopen() error for %s", ALMNAM, cmd2);
		free(cmd2);
		r = mystrdup(err, fnnam);
		return r;
	}
	if (fgets(buf, MAXBUF, infile) == NULL) {
		if (ferror(infile)) {
			logreturn(PIKTLOG, LOG_ERR, ERRID_FGETS_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, fgets() error for %s", ALMNAM, cmd2);
			free(cmd2);
			/* clearerr(infile); */
			inputclose(infile, pid);
			r = mystrdup(err, fnnam);
			return r;
		} else {
			free(cmd2);
			inputclose(infile, pid);
			r = mystrdup(nullstring, fnnam);
			return r;
		}
	}
	inputclose(infile, pid);
	free(cmd2);
	r = mystrdup(buf, fnnam);
	if ((i = dequote(r)) > 0) {
		if (*(r+i-1) == '\n')
			*(r+i-1) = '\0';
	}
	return r;
}

/* for a given file path, return the basename
   e.g., my_basename("/etc/mail/aliases") --> "aliases"
 */
char *
my_basename (char *path)
{
	static char fnnam[] = "my_basename";
	char *cp;
	char *r;

	if ((cp = strrchr(path, '/')) == NULL) {
		r = mystrdup(path, fnnam);
	} else {
		r = mystrdup(cp+1, fnnam);
	}
	return r;
}

/* for a given file path, return the directory portion
   e.g., my_dirname("/etc/mail/aliases") --> "/etc/mail/"
 */
char *
my_dirname (char *path)
{
	static char fnnam[] = "my_dirname";
  	char *cp;
	char *r;

	r = mystrdup(path, fnnam);
	if ((cp = strrchr(r, '/')) == NULL)
		*r = '\0';
	else if (*(cp+1))
		*(cp+1) = '\0';
	return r;
}

/* for the given path, return the file type and mode (permissions)
 */
char *
filemode (const char *path)
{
	static char fnnam[] = "filemode";
	struct stat fbuf;
	char *r;

	if (stat(path, &fbuf) != 0) {
		if (isverboselog)
			logreturn(PIKTLOG, LOG_WARNING, ERRID_FILE_STAT_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, stat() error for %s", ALMNAM, path);
		r = mystrdup(err, fnnam);
	} else {
		r = mymalloc(7, fnnam);
		sprintf(r, "%06o", fbuf.st_mode);
	}
	return r;
}

/* for the given path, return the i-node number
 */
int
fileino (const char *path)
{
	static char fnnam[] = "fileino";
	struct stat fbuf;

	if (stat(path, &fbuf) != 0) {
		if (isverboselog)
			logreturn(PIKTLOG, LOG_WARNING, ERRID_FILE_STAT_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, stat() error for %s", ALMNAM, path);
		return ERROR;
	} else {
		return (int) fbuf.st_ino;
	}
}

/* for the given path, return the number of links
 */
int
filelinks (const char *path)
{
	static char fnnam[] = "filelinks";
	struct stat fbuf;

	if (stat(path, &fbuf) != 0) {
		if (isverboselog)
			logreturn(PIKTLOG, LOG_WARNING, ERRID_FILE_STAT_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, stat() error for %s", ALMNAM, path);
		return ERROR;
	} else {
		return (int) fbuf.st_nlink;
	}
}

/* for the given path, return the uid
 */
int
fileuid (const char *path)
{
	static char fnnam[] = "fileuid";
	struct stat fbuf;

	if (stat(path, &fbuf) != 0) {
		if (isverboselog)
			logreturn(PIKTLOG, LOG_WARNING, ERRID_FILE_STAT_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, stat() error for %s", ALMNAM, path);
		return ERROR;
	} else {
		return (int) fbuf.st_uid;
	}
}

/* for the given path, return the gid
 */
int
filegid (const char *path)
{
	static char fnnam[] = "filegid";
	struct stat fbuf;

	if (stat(path, &fbuf) != 0) {
		if (isverboselog)
			logreturn(PIKTLOG, LOG_WARNING, ERRID_FILE_STAT_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, stat() error for %s", ALMNAM, path);
		return ERROR;
	} else {
		return (int) fbuf.st_gid;
	}
}

/* for the given path, return the size in bytes
 */
long
filesize (const char *path)
{
	static char fnnam[] = "filesize";
	struct stat fbuf;

	if (stat(path, &fbuf) != 0) {
		if (isverboselog)
			logreturn(PIKTLOG, LOG_WARNING, ERRID_FILE_STAT_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, stat() error for %s", ALMNAM, path);
		return ERROR;
	} else {
		return (long) fbuf.st_size;
	}
}

/* for the given path, return the time of last access
 */
time_t
fileatime (const char *path)
{
	static char fnnam[] = "fileatime";
	struct stat fbuf;

	if (stat(path, &fbuf) != 0) {
		if (isverboselog)
			logreturn(PIKTLOG, LOG_WARNING, ERRID_FILE_STAT_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, stat() error for %s", ALMNAM, path);
		return ERROR;
	} else {
		return fbuf.ST_ATIME;
	}
}

/* for the given path, return the time of last file status change
 */
time_t
filectime (const char *path)
{
	static char fnnam[] = "filectime";
	struct stat fbuf;

	if (stat(path, &fbuf) != 0) {
		if (isverboselog)
			logreturn(PIKTLOG, LOG_WARNING, ERRID_FILE_STAT_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, stat() error for %s", ALMNAM, path);
		return ERROR;
	} else {
		return fbuf.ST_CTIME;
	}
}

/* for the given path, return the time of last modification
 */
time_t
filemtime (const char *path)
{
	static char fnnam[] = "filemtime";
	struct stat fbuf;

	if (stat(path, &fbuf) != 0) {
		if (isverboselog)
			logreturn(PIKTLOG, LOG_WARNING, ERRID_FILE_STAT_ERROR, __FILE__, __LINE__, fnnam, TRUE, "in %s, stat() error for %s", ALMNAM, path);
		return ERROR;
	} else {
		return fbuf.ST_MTIME;
	}
}

/* for the given uid or name, return the name in the passwd file
 */
char *
pwname (int id, const signed char *name)
{
	static char fnnam[] = "pwname";
	struct passwd *pw;
	char *r;

	if ((id == NIL) && (*name != NIL))
		pw = getpwnam(name);
	else if ((id != NIL) && (*name == NIL))
		pw = getpwuid((uid_t) id);
	else {
		r = mystrdup(err, fnnam);
		return r;
	}
	if (pw == NULL)
		r = mystrdup(err, fnnam);
	else
		r = mystrdup(pw->pw_name, fnnam);
	return r;
}

/* for the given uid or name, return the passwd in the passwd file
 */
char *
pwpasswd (int id, const signed char *name)
{
	static char fnnam[] = "pwpasswd";
	struct passwd *pw;
	char *r;

	if ((id == NIL) && (*name != NIL))
		pw = getpwnam(name);
	else if ((id != NIL) && (*name == NIL))
		pw = getpwuid((uid_t) id);
	else {
		r = mystrdup(err, fnnam);
		return r;
	}
	if (pw == NULL)
		r = mystrdup(err, fnnam);
	else
		r = mystrdup(pw->pw_passwd, fnnam);
	return r;
}

/* for the given uid or name, return the uid in the passwd file
 */
int
pwuid (int id, const signed char *name)
{
	struct passwd *pw;

	if ((id == NIL) && (*name != NIL))
		pw = getpwnam(name);
	else if ((id != NIL) && (*name == NIL))
		pw = getpwuid((uid_t) id);
	else
		return ERROR;
	if (pw == NULL)
		return ERROR;
	else
		return pw->pw_uid;
}

/* for the given uid or name, return the gid in the passwd file
 */
int
pwgid (int id, const signed char *name)
{
	struct passwd *pw;

	if ((id == NIL) && (*name != NIL))
		pw = getpwnam(name);
	else if ((id != NIL) && (*name == NIL))
		pw = getpwuid((uid_t) id);
	else
		return ERROR;
	if (pw == NULL)
		return ERROR;
	else
		return pw->pw_gid;
}

/* for the given uid or name, return the gecos in the passwd file
 */
char *
pwgecos (int id, const signed char *name)
{
	static char fnnam[] = "pwgecos";
	struct passwd *pw;
	char *r;

	if ((id == NIL) && (*name != NIL))
		pw = getpwnam(name);
	else if ((id != NIL) && (*name == NIL))
		pw = getpwuid((uid_t) id);
	else {
		r = mystrdup(err, fnnam);
		return r;
	}
	if (pw == NULL)
		r = mystrdup(err, fnnam);
	else
		r = mystrdup(pw->pw_gecos, fnnam);
	return r;
}

/* for the given uid or name, return the dir in the passwd file
 */
char *
pwdir (int id, const signed char *name)
{
	static char fnnam[] = "pwdir";
	struct passwd *pw;
	char *r;

	if ((id == NIL) && (*name != NIL))
		pw = getpwnam(name);
	else if ((id != NIL) && (*name == NIL))
		pw = getpwuid((uid_t) id);
	else {
		r = mystrdup(err, fnnam);
		return r;
	}
	if (pw == NULL)
		r = mystrdup(err, fnnam);
	else
		r = mystrdup(pw->pw_dir, fnnam);
	return r;
}

/* for the given uid or name, return the shell in the passwd file
 */
char *
pwshell (int id, const signed char *name)
{
	static char fnnam[] = "pwshell";
	struct passwd *pw;
	char *r;

	if ((id == NIL) && (*name != NIL))
		pw = getpwnam(name);
	else if ((id != NIL) && (*name == NIL))
		pw = getpwuid((uid_t) id);
	else {
		r = mystrdup(err, fnnam);
		return r;
	}
	if (pw == NULL)
		r = mystrdup(err, fnnam);
	else
		r = mystrdup(pw->pw_shell, fnnam);
	return r;
}

/* for the given gid or name, return the name in the group file
 */
char *
grname (int id, const signed char *name)
{
	static char fnnam[] = "grname";
	struct group *gr;
	char *r;

	if ((id == NIL) && (*name != NIL))
		gr = getgrnam(name);
	else if ((id != NIL) && (*name == NIL))
		gr = getgrgid((gid_t) id);
	else {
		r = mystrdup(err, fnnam);
		return r;
	}
	if (gr == NULL)
		r = mystrdup(err, fnnam);
	else
		r = mystrdup(gr->gr_name, fnnam);
	return r;
}

/* for the given gid or name, return the passwd in the group file
 */
char *
grpasswd (int id, const signed char *name)
{
	static char fnnam[] = "grpasswd";
	struct group *gr;
	char *r;

	if ((id == NIL) && (*name != NIL))
		gr = getgrnam(name);
	else if ((id != NIL) && (*name == NIL))
		gr = getgrgid((gid_t) id);
	else {
		r = mystrdup(err, fnnam);
		return r;
	}
	if (gr == NULL)
		r = mystrdup(err, fnnam);
	else
		r = mystrdup(gr->gr_passwd, fnnam);
	return r;
}

/* for the given gid or name, return the gid in the group file
 */
int
grgid (int id, const signed char *name)
{
	struct group *gr;

	if ((id == NIL) && (*name != NIL))
		gr = getgrnam(name);
	else if ((id != NIL) && (*name == NIL))
		gr = getgrgid((gid_t) id);
	else
		return ERROR;
	if (gr == NULL)
		return ERROR;
	else
		return gr->gr_gid;
}

/* for the given gid or name, return the members in the group file
 */
char *
grmem (int id, const signed char *name)
{
	static char fnnam[] = "grmem";
	struct group *gr;
	char members[MAXLINE+1];
	int i;
	char *r;

	if ((id == NIL) && (*name != NIL))
		gr = getgrnam(name);
	else if ((id != NIL) && (*name == NIL))
		gr = getgrgid((gid_t) id);
	else {
		r = mystrdup(err, fnnam);
		return r;
	}
	if (gr == NULL)
		r = mystrdup(err, fnnam);
	else {
		members[0] = '\0';
		i = 0;
		while (*(gr->gr_mem+i) != NULL) {
			if (strlen(members) + strlen(*(gr->gr_mem+i)) >= MAXLINE) {
				logreturn(PIKTLOG, LOG_WARNING, ERRID_MEMBERS_TOO_MANY, __FILE__, __LINE__, fnnam, FALSE, "too many members in %s", gr->gr_name);
				break;
			}
			strcat(members, *(gr->gr_mem+i));
			if (*(gr->gr_mem+i+1) != NULL)
				strcat(members, ",");
			i++;
		}
		r = mystrdup(members, fnnam);
	}
	return r;
}

/* myrandom(-1) (or with any other negative arg) returns a random number
   0.0 <= r < 1.0;
   myrandom(n) (where n is a postive integer) returns a random integer
   from 1 to n;
   we might use random() instead of rand() here eventually, but rand()
   should be "good enough" for PIKT purposes
 */
double
myrandom(int n)
{
	double r;

	r = (double)rand()/((double)(RAND_MAX)+1.0);	/* random number, 0.0 <= r < 1.0 */
	if (n >= 0)
		r = (double) ((int)(r*n) +1);
	return r;
}
